unit ESBDates;

{: ESB Date/Time Routines Collection v 1.1

	Date/Time Routines to enhance your 32-bit Delphi
	Programming. <p>

	(c) 1999 ESB Consultancy <p>

	Recommend checking out (included): <p>
 <p>
		URL: http://www.pip.dknet.dk/~c-t/calendar.html <p>
 <p>
			    FREQUENTLY ASKED QUESTIONS ABOUT <p>
						CALENDARS <p>
 <p>
	Some of our algorithms are based on info from the above. <p>

 <p>

	These routines are used by ESB Consultancy within the
	development of their Customised Application. <p>

	ESB Consultancy retains full copyright. <p>

	ESB Consultancy grants users of this code royalty free rights
	to do with this code as they wish. <p>

	ESB Consultancy makes no guarantees nor excepts any liabilities
	due to the use of these routines. <p>

	We do ask that if this code helps you in you development
	that you send as an email mailto:info@esbconsult.com.au or even
	a local postcard. It would also be nice if you gave us a
	mention in your About Box, Help File or Documentation. <p>

	ESB Consultancy Home Page: http://www.esbconsult.com.au <p>

	Mail Address: PO Box 2259, Boulder, WA 6432 AUSTRALIA <p>

	History: <p>
<p>
	v 1.1 6-Apr-1999<p>
	- Added Better IsLeapYear<p>
	- Added GetFirstDayOfMonth & GetLastDayOfMonth<p>
	- Added Constants for Tropical Year and Synodic Month<p>
	- Added Calculation of GoldenNumber and Epact for given year<p>
	- Added GetGoodFriday and GetEasterSunday for given year<p>
	- Added GMTStr2Value<p>
	- Added GetLocalTZBias - which allows for Daylight Savings settings<p>
	- Added GetLocalTime - turns GMT into Local via Regional Settings<p>
	- All Local Time Zone routines handle Standard/Daylight Biases<p>
	- Added ConvertGMTStrTimes to convert between two GMT related Date/Times<p>
<p>
	v 1.0 6-Mar-1999 Intial Release <p>
}

interface

{.$Define UseESBRoutines} // Uncomment if you have ESB Routines

const
	{: Time it takes, in days, for the Earth to go from a point in
		its rotation and return to that point, i.e. a revolution.
		This amount changes (gets smaller) as time progresses.
		In 2100 it will be 365.242184 }
	TropicalYear = 365.242190;

	{: Time it takes, in days, from one New Moon to the next.
		This amount change (gets larger) as time progresses.
		in 2100 it will be 29.5305891 }
	SynodicMonth = 29.5305889;

//--- Year Based Routines ---

{: Is given Year a Leap Year. Thanks to Dr John Stockton
	[jrs@merlyn.demon.co.uk] for suggesting a faster methodology }
function IsLeapYear (Year: Word): Boolean;

{: Returns the GoldenNumber for a given Year. Values are 1 -> 19.
	The relationship between the Moon's Phases and the Year, repeats
	itself every 19 years. }
function GetGoldenNumber (const Year: Word): Integer;

{: Return the Epact, which is a measure of the age of the moon (i.e. the number
	of days that have passed since an "official" new moon) on a particular date.
	We have uses the Gregorian Calendar calculation }
function GetEpact (const Year: Word): Integer;

{:Returns the Date of Easter Sunday for given Year - based on current Calendar }
function GetEasterSunday (const Year: Word): TDateTime;

{:Returns the Date of Good Friday for given Year - based on current Calendar }
function GetGoodFriday (const Year: Word): TDateTime;

//--- Conversions ---

{: Returns Date as a String using ShortDateFormat from Regional
	Settings and an Empty String if Date < 1.0.
	If an error occurs an Empty String is Returned}
function Date2Str (const DT: TDateTime): String;

//: Returns the Day of the Month number from a given date/time
function Date2Day (const DT: TDateTime): Word;

//: Returns the Month number from a given date/time, 1 = Jan, ...
function Date2Month (const DT: TDateTime): Word;

//: Returns the Year from a given date/time
function Date2Year (const DT: TDateTime): Word;

{: Returns the Time Portion as a string HH:MM with time separator
	from the Regional Strings
	If an error occurs an Empty String is Returned}
function Time2Str (const DT: TDateTime): String;

//: Returns the Hour from a given date/time
function Time2Hr (const DT: TDateTime): Word;

//: Returns the Minute from a given date/time
function Time2Min (const DT: TDateTime): Word;

//: Returns the Second from a given date/time
function Time2Sec (const DT: TDateTime): Word;

//: Returns the Millisecond from a given date/time
function Time2MSec (const DT: TDateTime): Word;

{: Converts a string containing a Time into a DateTime}
function Str2Time (S: String): TDateTime;

{: Converts a string containing a Date into a DateTime}
function Str2Date (S: String): TDateTime;

//--- Date/Time Arithmetic ---

{: Returns the number of whole minutes apart the two times are (date portion ignored)
	and DT1 is assumed to be before DT2}
function MinutesApart (const DT1, DT2: TDateTime): Word;

{: Converts a time in MilliSeconds to a string of the form H:MM:SS.mmm }
function MS2TimeStr (const MS: LongInt): String;

{: Returns the current date/time as a string in the Format of: YYYYMMDD-HHMMSSmmm}
function GetDateTimeStamp: String;

{: Adjusts the date so that it has the Year specified. }
function AdjustDateYear (const D: TDateTime; const Year: Word): TDateTime;

{: Adds a Floating Point amount of Seconds to a Given Time
	Date portion preserved. }
function AddSecs (const DT: TDateTime; const Secs: Extended): TDateTime;

{: Adds a Floating Point amount of Minutes to a Given Time
	Date portion preserved. }
function AddMins (const DT: TDateTime; const Mins: Extended): TDateTime;

{: Adds a Floating Point amount of Hours to a Given Time
	Time portion preserved. }
function AddHrs (const DT: TDateTime; const Hrs: Extended): TDateTime;

{: Adds a Floating Point amount of Weeks to a Given Time
	Time portion preserved. }
function AddWeeks (const DT: TDateTime; const Weeks: Extended): TDateTime;

{: Adds a Floating Point amount of Fortnights to a Given Time
	Time portion preserved. }
function AddFortnights (const DT: TDateTime; const FNights: Extended): TDateTime;

{: Adds a Floating Point amount of Months to a Given Time
	Fractional portion of Month is assumed to be related to 30 day months.
	Time portion preserved. <p>
	<p>
	If adding Months results in landing on a nonsense date like 31 Apr
	then the last day in the month is used. This only applies to the
	integral component of the Months Added. The fractional part always
	is added to the resultant Date/Time }
function AddMonths (const DT: TDateTime; const Months: Extended): TDateTime;

{: Adds a Floating Point amount of Quarters to a Given Time
	Fractional portion of Quarter is assumed to be related to 30 day month,
	as the AddMonths routine is used.
	Time portion preserved. <p>
	<p>
	If adding Quarters results in landing on a nonsense date like 31 Apr
	then the last day in the month is used. This only applies to the
	integral component of the Quarter Added. The fractional part always
	is added to the resultant Date/Time }
function AddQuarters (const DT: TDateTime; const Qtrs: Extended): TDateTime;

{: Adds a Floating Point amount of Semesters to a Given Time
	Fractional portion of Semester is assumed to be related to 30 day month,
	as the AddMonths routine is used.
	Time portion preserved. <p>
	<p>
	If adding Semesters results in landing on a nonsense date like 31 Apr
	then the last day in the month is used. This only applies to the
	integral component of the Semeters Added. The fractional part always
	is added to the resultant Date/Time }
function AddSemesters (const DT: TDateTime; const Sems: Extended): TDateTime;

{: Adds a Floating Point amount of Years to a Given Time
	Fractional portion of Year is assumed to be related to 365.25 day years.
	Time portion preserved.<p>
	<p>
	If Adding Years results in landing on Feb 29 in a non-leap year, then
	this will be converted to Feb 28.This only applies to the
	integral component of the Years Added. The fractional part always
	is added to the resultant Date/Time }
function AddYrs (const DT: TDateTime; const Yrs: Extended): TDateTime;

{: Returns Last Day of the Month, for a given Date/Time - Time portion preserved }
function GetLastDayofMonth (const DT: TDateTime): TDateTime;

{: Returns First Day of the Month, for a given Date/Time - Time portion preserved }
function GetFirstDayofMonth (const DT: TDateTime): TDateTime;

//--- TimeZone Routines ---

{: Returns the Local TimeZone Bias, in Minutes, taking into account Standard
	and/or Daylight Biases, from the Regional Settings }
function GetLocalTZBias: LongInt;

{: Returns the TimeZone Name from the Regional Settings}
function GetTimeZoneName: string;

{: Returns the Difference from GMT (Greenwich Mean Time) as a string. For
	example mine is "GMT+8" }
function GetGMTDifference: string;

{: Returns the GMT string as a value. For example mine is "GMT+8" would return 8.0.
	Must contain a '+' or '-' sign, the 'GMT' is actually ignored currently,
	though this may change in the future }
function GMTStr2Value(const GMTStr: string): Extended;

{: Uses Regional Strings to determine your TimeZone and returns the
	Date/Time given (taken as being in your Time Zone) as GMT (Greenwich
	Mean Time) based }
function GetGMTTime (const DT: TDateTime): TDateTime;

{: Uses Regional Strings to determine your TimeZone and returns the
	Date/Time given (taken as being in GMT) as Local Time based }
function GetLocalTime (const DT: TDateTime): TDateTime;

{: Converts a Given Date/Time based on FromGMTStr (e.g. GMT+4) and returns
	a Date/Time using the ToGMTStr (e.g. GMT-5) }
function ConvertGMTStrTimes (const FromGMTStr: string; const FromDT: TDateTime;
	const ToGMTStr: string): TDateTime;

implementation

uses
	{$IFDEF UseESBRoutines}
	ESBRtns,
	{$ENDIF}
	Windows, SysUtils;

{$IFNDEF UseESBRoutines}
// --- Routines taken from ESBRoutines ---

function LInt2EStr (const L: LongInt): String;
begin
	try
		Result := IntToStr (L);
	except
		Result := '';
	end;
end;

function LeftStr (const S : string; const N : Integer): string;
begin
	Result := Copy (S, 1, N);
end;

function FillStr (const Ch : Char; const N : Integer): string;
begin
	SetLength (Result, N);
	FillChar (Result [1], N, Ch);
end;

function PadChLeftStr (const S : string; const Ch : Char;
	const Len : Integer): string;
var
	N: Integer;
begin
	N := Length (S);
	if N < Len then
		Result := FillStr (Ch, Len - N) + S
	else
		Result := S;
end;

function LInt2ZStr (const L: LongInt; const Len: Byte): String;
begin
	Result := LInt2EStr (L);
	Result := PadChLeftStr (LeftStr (Result, Len), '0', Len);
end;

function ReplaceChStr (const S : string;
	const OldCh, NewCh : Char): string;
var
	I: Integer;
begin
	Result := S;
	if OldCh = NewCh then
		Exit;
	for I := 1 to Length (S) do
		if S [I] = OldCh then
			Result [I] := NewCh;
end;

function Str2Ext (const S: String): Extended;
var
	S2: string;
begin
	try
		Result := StrToFloat (S2);
	except
		Result := 0;
	end;
end;
// --- End Routines taken from ESBRoutines ---
{$ENDIF}

function IsLeapYear (Year: Word): Boolean;
begin
	Result := ((Year and 3) = 0) and ((Year mod 100 > 0) or (Year mod 400 = 0))
end;

function Date2Str (const DT: TDateTime): String;
begin
	try
		if DT < 1.0 then
			Result := ''
		else
			Result := DateToStr (DT);
	except
		Result := '';
	end;
end;

function Time2Str (const DT: TDateTime): String;
var
	Hrs, Mins, Secs, MSecs: Word;
begin
	try
		DecodeTime (DT, Hrs, Mins, Secs, MSecs);
		Result := LInt2EStr (Hrs) + TimeSeparator
			+ LInt2ZStr (Mins, 2);
	except
		Result := '';
	end;
end;

function Date2Year (const DT: TDateTime): Word;
var
	D, M: Word;
begin
	DecodeDate (DT, Result, M, D);
end;

function Date2Month (const DT: TDateTime): Word;
var
	D, Y : Word;
begin
	DecodeDate (DT, Y, Result, D);
end;

function Date2Day (const DT: TDateTime): Word;
var
	M, Y : Word;
begin
	DecodeDate (DT, Y, M, Result);
end;

function Time2Hr (const DT: TDateTime): Word;
var
	Min, Sec, MSec: Word;
begin
	DecodeTime (DT, Result, Min, Sec, MSec);
end;

function Time2Min (const DT: TDateTime): Word;
var
	Hr, Sec, MSec: Word;
begin
	DecodeTime (DT, Hr, Result, Sec, MSec);
end;

function Time2Sec (const DT: TDateTime): Word;
var
	Hr, Min, MSec: Word;
begin
	DecodeTime (DT, Hr, Min, Result, MSec);
end;

function Time2MSec (const DT: TDateTime): Word;
var
	Hr, Min, Sec: Word;
begin
	DecodeTime (DT, Hr, Min, Sec, Result);
end;

function MinutesApart (const DT1, DT2: TDateTime): Word;
var
	Hr1, Min1, Sec1, MSec1: Word;
	Hr2, Min2, Sec2, MSec2: Word;
begin
	DecodeTime (DT1, Hr1, Min1, Sec1, MSec1);
	DecodeTime (DT2, Hr2, Min2, Sec2, MSec2);
	if Min2 < Min1 then
	begin
		Min2 := Min2 + 60;
		Dec (Hr2);
	end;
	if Hr1 > Hr2 then
		Hr2 := Hr2 + 24;
	Result := (Hr2 - Hr1) * 60 + (Min2 - Min1);
end;

function MS2TimeStr (const MS: LongInt): String;
var
	L: LongInt;
begin
	L := MS;
	Result := '.' + LInt2ZStr (L mod 1000, 3);
	L := L div 1000;
	Result := ':' + LInt2ZStr (L mod 60, 2) + Result;
	L := L div 60;
	Result := ':' + LInt2ZStr (L mod 60, 2) + Result;
	L := L div 60;
	Result := LInt2EStr (L) + Result ;
end;

function AdjustDateYear (const D: TDateTime; const Year: Word): TDateTime;
var
	Day, Month, OldYear: Word;
begin
	DecodeDate (D, OldYear, Month, Day);
	if Year = OldYear then
	begin
		Result := Int (D);
		Exit;
	end;
	if not IsLeapYear (Year) and (Month = 2) and (Day = 29) then
	begin
		Month := 3;
		Day := 1;
	end;
	Result := EncodeDate (Year, Month, Day);
end;

function GetDateTimeStamp: String;
var
	DT: TDateTime;
	Year, Month, Day: Word;
	Hr, Min, Sec, MSec: Word;
begin
	DT := Now;
	DecodeDate (DT, Year, Month, Day);
	DecodeTime (DT, Hr, Min, Sec, MSec);
	Result := LInt2ZStr (Year, 4) + LInt2ZStr (Month, 2) +
		LInt2ZStr (Day, 2) + '-' + LInt2ZStr (Hr, 2) +
		LInt2ZStr (Min, 2) + LInt2ZStr (Sec, 2) + Lint2ZStr (MSec, 3);
end;

function Str2Time (S: String): TDateTime;
begin
	try
		S := ReplaceChStr ( S, '.', TimeSeparator);
		S := ReplaceChStr ( S, ':', TimeSeparator);
		Result := Frac (StrToTime (S));
	except
		Result := 0.0;
	end;
end;

function Str2Date (S: String): TDateTime;
begin
	try
		S := ReplaceChStr ( S, '-', DateSeparator);
		S := ReplaceChStr ( S, '/', DateSeparator);
		Result := Trunc (StrToDate (S));
	except
		Result := 0.0;
	end;
end;

function AddSecs (const DT: TDateTime; const Secs: Extended): TDateTime;
begin
	Result := DT + Secs / (60 * 60 * 24)
end;

function AddMins (const DT: TDateTime; const Mins: Extended): TDateTime;
begin
	Result := DT + Mins / (60 * 24)
end;

function AddHrs (const DT: TDateTime; const Hrs: Extended): TDateTime;
begin
	Result := DT + Hrs / 24.0
end;

function AddWeeks (const DT: TDateTime; const Weeks: Extended): TDateTime;
begin
	Result := DT + Weeks * 7;
end;

function AddFortnights (const DT: TDateTime; const FNights: Extended): TDateTime;
begin
	Result := AddWeeks (DT, FNights * 2);
end;

function AddMonths (const DT: TDateTime; const Months: Extended): TDateTime;
var
	Day, Month, Year: Word;
begin
	DecodeDate (DT, Year, Month, Day);
	Month := Month + Trunc (Months);
	if Month > 12 then
	begin
		Year := Year + (Month - 1) div 12;
		Month := Month mod 12;
		if Month = 0 then
			Month := 12;
	end;

	// Ensure Day of Month is valid
	if Month = 2 then
	begin
		if IsLeapYear (Year) and (Day > 29) then
			Day := 29
		else	if not IsLeapYear (Year) and (Day > 28) then
			Day := 28;
	end
	else if (Month in [9, 4, 6, 11]) and (Day = 31) then
		Day := 30;

	Result := EncodeDate (Year, Month, Day) + Frac (Months) * 30 +
		Frac (DT);
end;

function AddQuarters (const DT: TDateTime; const Qtrs: Extended): TDateTime;
begin
	Result := AddMonths (DT, Qtrs * 3);
end;

function AddSemesters (const DT: TDateTime; const Sems: Extended): TDateTime;
begin
	Result := AddMonths (DT, Sems * 6);
end;

function AddYrs (const DT: TDateTime; const Yrs: Extended): TDateTime;
var
	Day, Month, Year: Word;
begin
	DecodeDate (DT, Year, Month, Day);
	Year := Year + Trunc (Yrs);
	if not IsLeapYear (Year) and (Month = 2) and (Day = 29) then
		Day := 28;
	Result := EncodeDate (Year, Month, Day) + Frac (Yrs) * 365.25
		+ Frac (DT);
end;

function GetLocalTZBias: LongInt;
var
	TZ : TTimeZoneInformation;
begin
	case GetTimeZoneInformation (TZ) of
		TIME_ZONE_ID_STANDARD: Result := TZ.Bias + TZ.StandardBias;
		TIME_ZONE_ID_DAYLIGHT: Result := TZ.Bias + TZ.DaylightBias;
	else
		Result := TZ.Bias;
	end;
end;

function GetGMTTime (const DT: TDateTime): TDateTime;
begin
	Result := DT + GetLocalTZBias / (24 * 60);
end;

function GetLocalTime (const DT: TDateTime): TDateTime;
begin
	Result := DT - GetLocalTZBias / (24 * 60);
end;

function GetTimeZoneName: string;
var
	TZ : TTimeZoneInformation;
begin
	case GetTimeZoneInformation (TZ) of
		TIME_ZONE_ID_STANDARD: Result := WideCharToString (TZ.StandardName);
		TIME_ZONE_ID_DAYLIGHT: Result := WideCharToString (TZ.DaylightName);
	else
		Result := '';
	end;
end;

function GetGMTDifference: string;
var
	TZ : TTimeZoneInformation;
begin
	GetTimeZoneInformation (TZ);
	if TZ.Bias = 0 then
		Result := 'GMT'
	else if TZ.Bias < 0 then
	begin
		if TZ.Bias mod 60 = 0 then
			Result := 'GMT+' + IntToStr (Abs (TZ.Bias) div 60)
		else
			Result := 'GMT+' + FloatToStr (Abs (TZ.Bias) / 60)
	end
	else
	begin
		if TZ.Bias mod 60 = 0 then
			Result := 'GMT-' + IntToStr (TZ.Bias div 60)
		else
			Result := 'GMT-' + FloatToStr (TZ.Bias / 60)
	end
end;

function GetLastDayofMonth (const DT: TDateTime): TDateTime;
var
	D, M, Y: Word;
begin
	DecodeDate (DT, Y, M, D);
	case M of
		2:
		begin
			if IsLeapYear (Y) then
				D := 29
			else
				D := 28;
		end;
		4, 6, 9, 11: D := 30
		else
			D := 31;
	end;
	Result := EncodeDate (Y, M, D) + Frac (DT);
end;

function GetFirstDayofMonth (const DT: TDateTime): TDateTime;
var
	D, M, Y: Word;
begin
	DecodeDate (DT, Y, M, D);
	Result := EncodeDate (Y, M, 1) + Frac (DT);
end;

function GetGoldenNumber (const Year: Word): Integer;
begin
	Result := Year mod 19 + 1;
end;

function GetEpact (const Year: Word): Integer;
var
	Century: Integer;
begin
	Century := Year div 100 + 1;
	Result := ((11 * (GetGoldenNumber (Year) - 1)) mod 30
		+ (8 * Century + 5) div 25) - (3 * Century) div 4  + 8;
	while Result < 1 do
		Result := Result + 30;
	while Result > 30 do
		Result := Result - 30;
end;

function GetEasterSunday (const Year: Word): TDateTime;
var
	C, I, J, H, G, L: Integer;
	D, M: Word;
begin
	G := GetGoldenNumber (Year) - 1;
	C := Year div 100;
	H := (C - C div 4 - (8 * C + 13) div 25 + 19 * G + 15) mod 30;
	I := H - (H div 28) * (1 - (H div 28) * (29 div (H + 1))*((21 - G) div 11));
	J := (Year + Year div 4 + I + 2 - C + C div 4) mod 7;

	L := I - J;
	M := 3 + (L + 40) div 44;
	D := L + 28 - 31 * (M div 4);
	Result := EncodeDate (Year, M, D);
end;

function GetGoodFriday (const Year: Word): TDateTime;
begin
	Result := GetEasterSunday (Year) - 2;
end;

function GMTStr2Value(const GMTStr: string): Extended;
var
	P: Integer;
begin
	P := Pos (GMTStr, '+');
	if P > 0 then
	begin
		Result := Str2Ext (Trim (Copy (GMTStr, P + 1, Length (GMTStr) - P)));
	end
	else
	begin
		P := Pos (GMTStr, '-');
		if P > 0 then
		begin
			Result := -1 * Str2Ext (Trim (Copy (GMTStr, P + 1, Length (GMTStr) - P)));
		end
		else
			Result := 0;
	end;
end;

function ConvertGMTStrTimes (const FromGMTStr: string; const FromDT: TDateTime;
	const ToGMTStr: string): TDateTime;
var
	GMT1, GMT2: Extended;
begin
	GMT1 := GMTStr2Value (FromGMTStr);
	GMT2 := GMTStr2Value (ToGMTStr);
	Result := FromDT + GMT2 - GMT1;
end;

end.

